/*
modification history
--------------------
01d,12dec07,kbm merges from 336 (FVS336G_POST_59_BUILD) branch done.
*/
#include <sys/mbuf.h>
#include <linux/rtnetlink.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <net/pfkeyv2.h> 
#include <netkey/keydb.h>
#include <netkey/key.h>
#include <keysock.h>
#include <netkey/key_debug.h>
#include <netinet6/ipsec.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <bufMgrLib.h>

/*
 * these are the heuristic values, derived considering sk_buff size needed for
 * dumping an SA. in worst case the sock recv and send queue can hold maxNoOf 
 * SAs.
 * for x86, the sk_buff size needed for dumping an SA: 424 bytes
 * for xscale, the sk_buff size needed for dumping an SA: 360 bytes.
 */

#define SA_DUMP_SIZE        (512)
extern int maxNoOfSa;
extern int keySockBuffSize;


struct sockaddr key_dst = {
            sa_family: PF_KEY,
};
struct sockaddr key_src = {
            sa_family: PF_KEY,
};

extern struct proto_ops key_raw_ops;

/** This hash table stores references to all the PF_KEY sockets **/
struct sock *key_sock_htable[MAX_KEY_PROTOS];
rwlock_t key_sock_htable_lock = RW_LOCK_UNLOCKED;

/* key_sock_hash(): inserts the socket into the hash table during creation
 */ 
static void key_sock_hash(struct sock *sk)
{
    struct sock **skp = &key_sock_htable[sk->sk_protocol & (MAX_KEY_PROTOS-1)];
    write_lock_bh(&key_sock_htable_lock);
    if ((socktokeycb(sk)->sk_next = *skp) != NULL)
        socktokeycb(*skp)->sk_pprev = &(socktokeycb(sk)->sk_next);
    *skp = sk;
    socktokeycb(sk)->sk_pprev = skp;
    sock_hold(sk);
    write_unlock_bh(&key_sock_htable_lock);
}

/* key_sock_unhash(): removes the socket from the hash table during close
 */ 
static void key_sock_unhash(struct sock *sk)
{
    write_lock_bh(&key_sock_htable_lock);
    if (socktokeycb(sk)->sk_pprev) {
        struct sock * next_sock = socktokeycb(sk)->sk_next;
        struct sock ** pprev_sock = socktokeycb(sk)->sk_pprev;
        if (next_sock)
            socktokeycb(next_sock)->sk_pprev = pprev_sock;
        *pprev_sock = next_sock;
        socktokeycb(sk)->sk_pprev = NULL;
        socktokeycb(sk)->sk_next = NULL;
        __sock_put(sk);
    }
    write_unlock_bh(&key_sock_htable_lock);
}

/* one of the 2 possible PF_KEY sub-protocol types */
struct proto pfkeyv2_prot = {
    name:           "PF_KEY_V2",
    hash:           key_sock_hash,
    unhash:         key_sock_unhash,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
    slab:	    NULL,	    
    obj_size:       sizeof(struct sock),		    
#endif
};

/** called by key_release **/
void key_sock_release(struct sock *sk)
{
    /* Remove from hash table */
    sk->sk_prot->unhash(sk);

    sock_orphan(sk);
    sock_put(sk);
}
#ifdef FIMPLE
/* redundancy chk */
#endif
void key_sock_destruct(struct sock* sk)
{
    __skb_queue_purge(&sk->sk_receive_queue);

    if (!sock_flag(sk, SOCK_DEAD)) {
        ipsec_print("Attempt to release alive pfkey socket %p\n", sk);
        return;
    }

    BUG_TRAP(atomic_read(&sk->sk_rmem_alloc) == 0);
    BUG_TRAP(atomic_read(&sk->sk_wmem_alloc) == 0);
    BUG_TRAP(sk->sk_wmem_queued == 0);
    BUG_TRAP(sk->sk_forward_alloc == 0);

    dst_release(sk->sk_dst_cache);/** sk->dst_cache is prolly NULL **/
    if (sk->sk_protinfo)
        bufFree(sk->sk_protinfo);
    sk->sk_protinfo = NULL;
}

/* key_create():called when userland creates a socket.
 *      It does nominal socket creation and intialization. 
 *      protocol must be 2(1 is not supported) 
 */
int key_create(struct socket* sock,int protocol)
{
        struct sock* sk;

    KEYDEBUG(KEYDEBUG_KEY_STAMP,
        ipsec_print("key_create(): PF_key: Setting up socket\n"));
                
    /* state is SS_CONNECTED at the end of this routine
     * after passing all failure cases below
     */
    if(sock->type != SOCK_RAW || protocol != 2)
        return -EINVAL;
    sock->state = SS_UNCONNECTED;
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
    sk = sk_alloc(PF_KEY, GFP_KERNEL, 1, NULL);
#else
    sk=sk_alloc(PF_KEY,GFP_KERNEL,&pfkeyv2_prot,1);
#endif
    if (sk == NULL)
        return -ENOBUFS;
    sk->sk_protinfo = bufAlloc(sizeof(struct keycb), GFP_KERNEL);
    if (sk->sk_protinfo == NULL)
    {
        sk_free(sk);
        return -ENOBUFS;
    }
    sock->ops = &key_raw_ops;
    sock_init_data(sock,sk);
    sk->sk_rcvbuf = (maxNoOfSa * SA_DUMP_SIZE);
    sk->sk_sndbuf = (maxNoOfSa * SA_DUMP_SIZE);
    keySockBuffSize = (maxNoOfSa * SA_DUMP_SIZE);
    sk->sk_prot = &pfkeyv2_prot;
    sk->sk_family = PF_KEY;
    sk->sk_protocol = protocol;
    /* sk->"callbacks"  set the call back functions
     */
    sk->sk_destruct = key_sock_destruct;
    
    /* sk->protoinfo.af_unix is used to store keycb(cb = control block)
     */
    {
        struct keycb *kp;       
        int af;
        kp = (struct keycb *)sk->sk_protinfo;
        kp->kp_promisc = kp->kp_registered = 0;
        af = sk->sk_family;
            if (af == PF_KEY) /*  AF_KEY */
            key_cb.key_count++;
        key_cb.any_count++;
        kp->kp_raw.rcb_laddr = &key_src;
        kp->kp_raw.rcb_faddr = &key_dst;
    }

    /* Add to protocol hash table */
    sk->sk_prot->hash(sk);
    
    sock->state = SS_CONNECTED;
    KEYDEBUG(KEYDEBUG_KEY_STAMP,
        ipsec_print("key_create(): PF_key: Socket created\n"));
    return 0;
}

/* key_sendmsg(): called when userland does a send.
 *      It creates an mbuf from the user iovec and passes it onto 
 *      key_output() for further processing. 
 */
int key_sendmsg(struct kiocb * iocb, struct socket *sock, 
        struct msghdr *msg, size_t total_len)
{
        struct iovec* iov;
    void *usr_data;
    size_t usr_data_len;
    unsigned flags = msg->msg_flags;
    struct mbuf *mb;
    
    KEYDEBUG(KEYDEBUG_KEY_DATA,
        ipsec_print("key_sendmsg(): PF_key: Sending message ->\n");
        kdebug_flag(flags));
    iov = msg->msg_iov;
    usr_data = iov->iov_base;
    usr_data_len = iov->iov_len;

    if(usr_data_len > MCLBYTES)
        return -EMSGSIZE; 

    /* Create a clustered mbuf out of the user data 
     * in msghdr and pass it to key_output 
     */
    MGETHDR(mb, (flags & MSG_DONTWAIT) ? M_NOWAIT : M_WAITOK, MT_DATA); 
    if (mb) {
        if(usr_data_len > MHLEN) {
            MCLGET(mb, (flags & MSG_DONTWAIT) ? M_NOWAIT : M_WAITOK);
            if ((mb->m_flags & M_EXT) == 0) {/* Ext mem alloc failed */
                m_free(mb);
                mb = NULL;
                return -ENOBUFS;
            }
        }
    }
    else 
        return -ENOBUFS;
    memcpy(mb->m_data, usr_data, usr_data_len);
    mb->m_pkthdr.len = mb->m_len = usr_data_len;
    mb->m_pkthdr.rcvif = NULL;
    mb->m_next = NULL;
    KEYDEBUG(KEYDEBUG_KEY_DUMP,
        ipsec_print("key_send(): PF_key ->\n");
        kdebug_mbuf(mb));
    return key_output(mb,sock);/* takes care of freeing mb */
}

/* key_recvmsg(): called when userland does a recv. 
 *      It uses skb_recv_datagram() to pickup sk_buffs from the 
 *      socket's receive queue. Then data is transferred from 
 *      sk_buff to user iovec and control is transferred to 
 *      user process.
 */
int key_recvmsg(struct kiocb * iocb, struct socket *sock, struct msghdr *msg, 
        size_t total_len, int flags)
{
    struct sock *sk = sock->sk;
    int addr_len = 0;
        struct sockaddr *skey = (struct sockaddr *)msg->msg_name;
    struct sk_buff *skb;
        int copied = 0;
    int error = -EOPNOTSUPP;
    int noblock = flags & MSG_DONTWAIT;

    KEYDEBUG(KEYDEBUG_KEY_DATA,
        ipsec_print("key_recvmsg(): PF_key: Receiving message -> \n");
        kdebug_flag(flags));
    if (flags & MSG_OOB)
        goto out;

    addr_len = sizeof(*skey);

    /* Pickup packet from socket buffer */
    skb = skb_recv_datagram(sk, flags, noblock, &error);
    if (!skb)
        goto out;

    KEYDEBUG(KEYDEBUG_KEY_DUMP,
        ipsec_print("key_recvmsg(): PF_key ->\n");
        kdebug_skbuff(skb));
    copied = skb->len;
    if (total_len < copied) {
        msg->msg_flags |= MSG_TRUNC;
        copied = total_len;
    }

    /* Copy packet to userland */
        error = skb_copy_datagram_iovec(skb, 0, msg->msg_iov, copied);
    if (error)
        goto done;
    sock_recv_timestamp(msg, sk, skb);

    /** set family **/
    if (skey) {
        skey->sa_family = PF_KEY;
        skey->sa_len = addr_len;
        msg->msg_namelen = addr_len;
    }
done:
    skb_free_datagram(sk, skb);
out:    
    error = error ? : copied;
    return error;
}

/* key_release():  called when userland does a close **/
int key_release(struct socket *sock)
{
    struct sock *sk = sock->sk;

    if (sk) {
        sock->sk = NULL;
        key_sock_release(sk);
    }
    key_freereg(sock);
    return(0);
}

/* key_poll:  called when userland does a select involving a 
 * pfkey socket descriptor 
 */
unsigned int key_poll(struct file * file, struct socket *sock, poll_table *wait)
{
    unsigned int mask;
    struct sock *sk = sock->sk;

    poll_wait(file, sk->sk_sleep, wait);
    mask = 0;

    /** exceptional events? **/
    if(sk->sk_err)
        mask |= POLLERR;
    if(sk->sk_shutdown == SHUTDOWN_MASK)
        mask |= POLLHUP;

    /** readable? **/
    if (!skb_queue_empty(&sk->sk_receive_queue) || 
        (sk->sk_shutdown&RCV_SHUTDOWN))
        mask |= POLLIN | POLLRDNORM;
    
    /** writeable? **/
    /** The current "socket write()" call results in the 
     ** user data being copied to a new mbuf everytime 
     ** Since "sk->sndbuf" is not used to pick data from user, 
     ** so it can never get full and hence can never block a write 
     ** for "buffer-space reasons". So pfkey sockets are always 
     ** writeable
     **/
    mask |= POLLOUT | POLLWRNORM;

    /** NB: POLLPRI represents priority data to be read and is used 
     ** when select is called with exception descriptor set. pfkey 
     ** doesnt set this since it doesnt have priority data. So select 
     ** on pfkey socket will never succeed when used with exception 
     ** descriptor set.
     **/

    return mask;
}

struct proto_ops key_raw_ops = {
        family:     PF_KEY,
    release:    key_release,
    bind:       sock_no_bind,
        connect:    sock_no_connect,
    socketpair:     sock_no_socketpair,
    accept:     sock_no_accept,
    getname:    sock_no_getname, 
    poll:       key_poll,
        ioctl:      sock_no_ioctl,
        listen:     sock_no_listen,
    shutdown:   sock_no_shutdown,
    setsockopt:     sock_no_setsockopt,
    getsockopt:     sock_no_getsockopt,
    sendmsg:    key_sendmsg,
    recvmsg:    key_recvmsg,
    mmap:       sock_no_mmap,
    sendpage:   sock_no_sendpage,
};

struct net_proto_family key_family_ops = {
        family: PF_KEY,
    create: key_create,
};

/* get_pfkey_sock(): called by pfkey_get_info */
static void get_pfkey_sock(struct sock *sk, char *tmpbuf, int i)
{
#ifdef FIMPLE
    /* redundancy check */
#endif
    sprintf(tmpbuf, 
        "%2d: "     /* hash number @i */
        "%3d  "     /* promiscous? */
        "%3d  "     /* registered? */
        "%02X",     /* socket state */
        i, 
        ((struct keycb *)socktorawcb(sk))->kp_promisc,
        ((struct keycb *)socktorawcb(sk))->kp_registered,
        sk->sk_state);
}

#define BUFSIZE 25
/* pfkey_get_info(): routine to view /proc entry for pfkey
 * @buffer : location(a page actually) to which the pfkey socket info is written
 * @start : this is location inside the page. It is used to indicate the exact 
 *      place where the socket info is written to.
 * @offset : indicates how far the buffer(=file=page) must be left untouched.
 * @length : indicates maximum allowable length of data that can be written
 *       with some slack.
 */
int pfkey_get_info(char *buffer, char **start, off_t offset, int length)
{
#ifdef FIMPLE
    /** put appropriate info in the proc file **/
#endif
    int len = 0, num = 0, i;
    /* @len will hold the value of number of bytes written to buffer */
    off_t pos = BUFSIZE;
    /* @pos indicates how far(from start of @buffer) the data may have been 
     * written. Since data can onyl be written after @offset(from start of 
     * @buffer). So @pos < @offset means we need to traverse further down the 
     * @buffer, while @pos > @offset means we can write @pos - @offset data.
     * So more often @len == @pos - @offset. But it is possible that 
     * @len != @pos - @offset when offset is not a multiple of BUFSIZE.
     */ 
    off_t begin = 0;
    /* @begin will be non-zero when len != pos - offset and used to set 
     * @start which will become different from @buffer
     */ 
    char tmpbuf[BUFSIZE+1];

    /* first call to this routine */
    if (offset < BUFSIZE)
        len += sprintf(buffer, "%-*s\n",BUFSIZE-1,"sl: prom regd st");
    read_lock(&key_sock_htable_lock);
    for (i = 0; i < MAX_KEY_PROTOS; i++) {
        struct sock *sk;

        for (sk = key_sock_htable[i]; sk; 
             sk = socktokeycb(sk)->sk_next, num++) {
            if(len + BUFSIZE >= length)
                goto out;
            pos += BUFSIZE;
            if (pos <= offset)
                continue;
            get_pfkey_sock(sk, tmpbuf, i);
            len += sprintf(buffer + len, "%-*s\n",BUFSIZE-1, tmpbuf);
        }
    }
out:
    read_unlock(&key_sock_htable_lock);
    begin = len - (pos - offset);
    *start = buffer + begin;
    len -= begin;
    if (len > length)
        len = length;
    if (len < 0)
        len = 0;
    return len;
}
